/*
 * Single-File-Header for I2C generic device
 * 05-07-2023 E. Brombaugh  / 21-11-2023 Pancio - atomic i2c ataptation
 */

#ifndef _I2C_H
#define _I2C_H

#include <string.h>

// DEVICE I2C address
// Should be defined in particular device's library
//#define I2C_ADDR 0x3c

//#ifndef SI5351_ADDRESS
//  #define SI5351_ADDRESS 0x72 // 0b11000000 for my module. Others may vary! The 0x60 did NOT work with my module!
//#endif

#define I2C_CLKRATE 100000  // I2C Bus clock rate - must be lower the Logic clock rate   
#define I2C_PRERATE 200000  // I2C Logic clock rate - must be higher than Bus clock rate
#define I2C_DUTY            // uncomment this for high-speed 36% duty cycle, otherwise 33%
#define TIMEOUT_MAX 100000  // I2C Timeout count
//#define I2C_IRQ             // uncomment this to enable IRQ-driven operation               

#ifdef I2C_IRQ
// some stuff that IRQ mode needs
volatile uint8_t i2c_send_buffer[64], *i2c_send_ptr, i2c_send_sz, i2c_irq_state;

//#define IRQ_DIAG          // uncomment this to enable time diags in IRQ
#endif

/*
 * init I2C
 */
void i2c_setup(void)
{
	uint16_t tempreg;
	
	// Reset I2C1 to init all regs
	RCC->APB1PRSTR |= RCC_APB1Periph_I2C1;
	RCC->APB1PRSTR &= ~RCC_APB1Periph_I2C1;
	
	// set freq
	tempreg = I2C1->CTLR2;
	tempreg &= ~I2C_CTLR2_FREQ;
	tempreg |= (FUNCONF_SYSTEM_CORE_CLOCK/I2C_PRERATE)&I2C_CTLR2_FREQ;
	I2C1->CTLR2 = tempreg;
	
	// Set clock config
	tempreg = 0;
#if (I2C_CLKRATE <= 100000)
	// standard mode good to 100kHz
	tempreg = (FUNCONF_SYSTEM_CORE_CLOCK/(2*I2C_CLKRATE))&I2C_CKCFGR_CCR;
#else
	// fast mode over 100kHz
#ifndef I2C_DUTY
	// 33% duty cycle
	tempreg = (FUNCONF_SYSTEM_CORE_CLOCK/(3*I2C_CLKRATE))&I2C_CKCFGR_CCR;
#else
	// 36% duty cycle
	tempreg = (FUNCONF_SYSTEM_CORE_CLOCK/(25*I2C_CLKRATE))&I2C_CKCFGR_CCR;
	tempreg |= I2C_CKCFGR_DUTY;
#endif
	tempreg |= I2C_CKCFGR_FS;
#endif
	I2C1->CKCFGR = tempreg;

#ifdef I2C_IRQ
	// enable IRQ driven operation
	NVIC_EnableIRQ(I2C1_EV_IRQn);
	
	// initialize the state
	i2c_irq_state = 0;
#endif
	
	// Enable I2C
	I2C1->CTLR1 |= I2C_CTLR1_PE;

	// set ACK mode
	I2C1->CTLR1 |= I2C_CTLR1_ACK;
}

/*
 * error descriptions
 */
char *errstr[] =
{
	"not busy",
	"master mode",
	"transmit mode",
	"tx empty",
	"transmit complete",
};

/*
 * error handler
 */
uint8_t i2c_error(uint8_t err)
{
	// report error
	printf("i2c_error - timeout waiting for %s\n\r", errstr[err]);
	
	// reset & initialize I2C
	i2c_setup();

	return 1;
}

// event codes we use
#define  I2C_EVENT_MASTER_MODE_SELECT ((uint32_t)0x00030001)                /* BUSY, MSL and SB flag */
#define  I2C_EVENT_MASTER_TRANSMITTER_MODE_SELECTED ((uint32_t)0x00070082)  /* BUSY, MSL, TXE, ADDR and TRA flags */
#define  I2C_EVENT_MASTER_BYTE_TRANSMITTED ((uint32_t)0x00070084)           /* BUSY, MSL, TXE, TRA and BTF flags */

/*
 * check for 32-bit event codes
 */
uint8_t i2c_chk_evt(uint32_t event_mask)
{
	/* read order matters here! STAR1 before STAR2!! */
	uint32_t status = I2C1->STAR1 | (I2C1->STAR2<<16);
	return (status & event_mask) == event_mask;
}

#ifdef I2C_IRQ
/*
 * packet send for IRQ-driven operation
 */
uint8_t i2c_send(uint8_t addr, uint8_t *data, uint8_t sz)
{
	int32_t timeout;
	
#ifdef IRQ_DIAG
	GPIOC->BSHR = (1<<(3));
#endif
	
	// error out if buffer under/overflow
	if((sz > sizeof(i2c_send_buffer)) || !sz)
		return 2;
	
	// wait for previous packet to finish
	while(i2c_irq_state);
	
#ifdef IRQ_DIAG
	GPIOC->BSHR = (1<<(16+3));
	GPIOC->BSHR = (1<<(4));
#endif
	
	// init buffer for sending
	i2c_send_sz = sz;
	i2c_send_ptr = i2c_send_buffer;
	memcpy((uint8_t *)i2c_send_buffer, data, sz);
	
	// wait for not busy
	timeout = TIMEOUT_MAX;
	while((I2C1->STAR2 & I2C_STAR2_BUSY) && (timeout--));
	if(timeout==-1)
		return i2c_error(0);

	// Set START condition
	I2C1->CTLR1 |= I2C_CTLR1_START;

	// wait for master mode select
	timeout = TIMEOUT_MAX;
	while((!i2c_chk_evt(I2C_EVENT_MASTER_MODE_SELECT)) && (timeout--));
	if(timeout==-1)
		return i2c_error(1);
	
	// send 7-bit address + write flag
	I2C1->DATAR = addr<<1;

	// wait for transmit condition
	timeout = TIMEOUT_MAX;
	while((!i2c_chk_evt(I2C_EVENT_MASTER_TRANSMITTER_MODE_SELECTED)) && (timeout--));
	if(timeout==-1)
		return i2c_error(2);

	// Enable TXE interrupt
	I2C1->CTLR2 |= I2C_CTLR2_ITBUFEN | I2C_CTLR2_ITEVTEN;
	i2c_irq_state = 1;

#ifdef IRQ_DIAG
	GPIOC->BSHR = (1<<(16+4));
#endif
	
	// exit
	return 0;
}


/*
 * IRQ handler for I2C events
 */
void I2C1_EV_IRQHandler(void) __attribute__((interrupt));
void I2C1_EV_IRQHandler(void)
{
	uint16_t STAR1, STAR2 __attribute__((unused));
	
#ifdef IRQ_DIAG
	GPIOC->BSHR = (1<<(4));
#endif

	// read status, clear any events
	STAR1 = I2C1->STAR1;
	STAR2 = I2C1->STAR2;
	
	/* check for TXE */
	if(STAR1 & I2C_STAR1_TXE)
	{
		/* check for remaining data */
		if(i2c_send_sz--)
			I2C1->DATAR = *i2c_send_ptr++;

		/* was that the last byte? */
		if(!i2c_send_sz)
		{
			// disable TXE interrupt
			I2C1->CTLR2 &= ~(I2C_CTLR2_ITBUFEN | I2C_CTLR2_ITEVTEN);
			
			// reset IRQ state
			i2c_irq_state = 0;
			
			// wait for tx complete
			while(!i2c_chk_evt(I2C_EVENT_MASTER_BYTE_TRANSMITTED));

			// set STOP condition
			I2C1->CTLR1 |= I2C_CTLR1_STOP;
		}
	}

#ifdef IRQ_DIAG
	GPIOC->BSHR = (1<<(16+4));
#endif
}
#else
//--------------------------------------------------------------------
/*
 * low-level START condition on i2c
 */
uint8_t i2c_start()
{
	int32_t timeout;
    
    // wait for not busy
	timeout = TIMEOUT_MAX;
	while((I2C1->STAR2 & I2C_STAR2_BUSY) && (timeout--));
	if(timeout==-1) return i2c_error(0);

	// Set START condition
	I2C1->CTLR1 |= I2C_CTLR1_START;
	
    return 0;
}  

/*
 * low-level send 7bit ADDRESS + READ/WRITE bit via i2c
 */
uint8_t i2c_send_addr(uint8_t data, uint8_t I2C_Direction)
{
	int32_t timeout;
    
    // wait for master mode select
	timeout = TIMEOUT_MAX;
	while((!i2c_chk_evt(I2C_EVENT_MASTER_MODE_SELECT)) && (timeout--));
	if(timeout==-1) return i2c_error(1);

    int8_t addr = data<<1;
    I2C1->DATAR = addr |= I2C_Direction; 
    
	// wait for transmit condition
	timeout = TIMEOUT_MAX;
	while((!i2c_chk_evt(I2C_EVENT_MASTER_TRANSMITTER_MODE_SELECTED)) && (timeout--));
	if(timeout==-1) return i2c_error(2); 
	
    return 0;    
}

/*
 * low-level send byte via i2c
 */
uint8_t i2c_send_byte(uint8_t data)
{
	int32_t timeout;
    
    // wait for TX Empty
	timeout = TIMEOUT_MAX;
	while(!(I2C1->STAR1 & I2C_STAR1_TXE) && (timeout--));
	if(timeout==-1) return i2c_error(3);
    
    I2C1->DATAR = data;
	return 0;    
}

/*
 * low-level stop condition on i2c
 */
uint8_t i2c_stop()
{
	int32_t timeout;
    
	// wait for tx complete
	timeout = TIMEOUT_MAX;
	while((!i2c_chk_evt(I2C_EVENT_MASTER_BYTE_TRANSMITTED)) && (timeout--));
	if(timeout==-1) return i2c_error(4);

	// set STOP condition
	I2C1->CTLR1 |= I2C_CTLR1_STOP;	
	return 0;
}

/*
 * low-level register/data pair send for blocking polled operation via i2c
 */
uint8_t i2c_send_datareg(uint8_t addr, uint8_t reg, uint8_t data)
{
	int32_t timeout;
	
	// wait for not busy
	timeout = TIMEOUT_MAX;
	while((I2C1->STAR2 & I2C_STAR2_BUSY) && (timeout--));
	if(timeout==-1) return i2c_error(0);

	// Set START condition
	I2C1->CTLR1 |= I2C_CTLR1_START;
	
    // wait for master mode select
	timeout = TIMEOUT_MAX;
	while((!i2c_chk_evt(I2C_EVENT_MASTER_MODE_SELECT)) && (timeout--));
	if(timeout==-1) return i2c_error(1);

    //int8_t addr = data<<1;
    I2C1->DATAR = addr<<1; 
    
	// wait for transmit condition
	timeout = TIMEOUT_MAX;
	while((!i2c_chk_evt(I2C_EVENT_MASTER_TRANSMITTER_MODE_SELECTED)) && (timeout--));
	if(timeout==-1) return i2c_error(2);
    
    // wait for TX Empty
	timeout = TIMEOUT_MAX;
	while(!(I2C1->STAR1 & I2C_STAR1_TXE) && (timeout--));
	if(timeout==-1) return i2c_error(3);
    
    I2C1->DATAR = reg; 
    
    // wait for TX Empty
	timeout = TIMEOUT_MAX;
	while(!(I2C1->STAR1 & I2C_STAR1_TXE) && (timeout--));
	if(timeout==-1) return i2c_error(3);
    
    I2C1->DATAR = data;

	// wait for tx complete
	timeout = TIMEOUT_MAX;
	while((!i2c_chk_evt(I2C_EVENT_MASTER_BYTE_TRANSMITTED)) && (timeout--));
	if(timeout==-1) return i2c_error(4);

	// set STOP condition
	I2C1->CTLR1 |= I2C_CTLR1_STOP;	
	return 0;    
    
}

//------------------------------------------------------------------


#endif

/*
 * init I2C and GPIO
 */
uint8_t i2c_init(void)
{
	// Enable GPIOC and I2C
	RCC->APB2PCENR |= RCC_APB2Periph_GPIOC;
	RCC->APB1PCENR |= RCC_APB1Periph_I2C1;
	
	// PC1 is SDA, 10MHz Output, alt func, open-drain
	GPIOC->CFGLR &= ~(0xf<<(4*1));
	GPIOC->CFGLR |= (GPIO_Speed_10MHz | GPIO_CNF_OUT_OD_AF)<<(4*1);
	
	// PC2 is SCL, 10MHz Output, alt func, open-drain
	GPIOC->CFGLR &= ~(0xf<<(4*2));
	GPIOC->CFGLR |= (GPIO_Speed_10MHz | GPIO_CNF_OUT_OD_AF)<<(4*2);
	
#ifdef IRQ_DIAG
	// GPIO diags on PC3/PC4
	GPIOC->CFGLR &= ~(0xf<<(4*3));
	GPIOC->CFGLR |= (GPIO_Speed_10MHz | GPIO_CNF_OUT_PP)<<(4*3);
	GPIOC->BSHR = (1<<(16+3));
	GPIOC->CFGLR &= ~(0xf<<(4*4));
	GPIOC->CFGLR |= (GPIO_Speed_10MHz | GPIO_CNF_OUT_PP)<<(4*4);
	GPIOC->BSHR = (1<<(16+4));
#endif

	// load I2C regs
	i2c_setup();
	
#if 0
	// test if I2C device is on the bus by sending command
	//uint8_t command = 0xAF;
	//return i2c_pkt_send(&command, 1, 1);
#else
	return 0;
#endif
}


/*
 * reset placeholder
 */
void rst(void)
{
}
#endif
